/**
 * @file capture.c
 * @brief This code gets images from an video interface using V4L2 
 * @author Florian Schmitt (0schmitt@informatik.uni-hamburg.de)
 * @date Jul-2012
**/
#include "camcapture.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <assert.h>

/* Device control headers */
#include <fcntl.h>
#include <unistd.h>
#include <sys/ioctl.h>

/* NMap (memory mapping) */
#include <sys/mman.h>

/* Video4Linux headers */
#include <asm/types.h>
#include <linux/videodev2.h>


/**
 * @struct buffer
 * This struct contains image information as a combination of its adress and size
 * Only use it in capture.c (private)
**/
struct buffer {
	void * start;   /**< Start position inside the image buffer */
	size_t length;  /**< Size of the image in bytes */
};


/* Globals */
const char *dev_name = NULL;         /**< device name (e.g. /dev/video0) */
struct v4l2_format fmt;              /**< video format information */
struct v4l2_buffer v4lBuf;           /**< video4linux buffer (NOT equals framebuffer!) */
struct buffer *buffers = NULL;       /**< framebuffer */
static unsigned int n_buffers = 0;   /**< count variable for frame buffers */
int fd;                              /**< filedescriptor for opened device */
enum v4l2_buf_type type;             /**< buffer type information for v4l */
int init_width;                      /**< init width */
int init_height;                     /**< init height */




/**
 * Does an IO-control as often as it needs to work
 * or an error occours
 * @param fd File descriptor of the device
 * @param request Requesttype ID
 * @param arg Arguments for the request (pointer)
 * @return r Success code
**/
int xioctl(int fd, unsigned long request, void *arg)
{
	int r;
	do r = ioctl(fd, request, arg);
	while((r == -1) && (EINTR == errno));
	return r;
}



/**
 * Initiate the video device to read from it
 * @param param_dev_name The device's name (e.g. /dev/video0)
 * @param width Image width
 * @param height Image height (wrong values may cause format errors)
 * @return success -> 1  /  failed -> 0
**/
int camcapture_init(const char *param_dev_name, unsigned int width, unsigned int height)
{
	struct v4l2_capability cap;
	struct v4l2_cropcap cropcap;
	struct v4l2_crop crop;
	struct v4l2_requestbuffers req;
	unsigned int min;

	dev_name = param_dev_name;
	init_width  = width;
	init_height = height;

	/* Opening Video device */
	fd = open(dev_name, O_RDWR | O_NONBLOCK, 0);
	if(-1 == fd)
	{
		fprintf(stderr, "Cannot open '%s'\n", dev_name);
		return 0;
	}

	/* Get the devices capabilities (cap) */
	if(xioctl(fd, VIDIOC_QUERYCAP, &cap) == -1) 
	{
		fprintf(stderr, "%s is no V4L2 device\n", dev_name);
		return 0;
	}

	/* Check if device is a video device (cap) */
	if(!(cap.capabilities & V4L2_CAP_VIDEO_CAPTURE)) {
		fprintf(stderr, "%s is no video capture device\n", dev_name);
		return 0;
	}

	/* Check streaming support (cap) */
	if(!(cap.capabilities & V4L2_CAP_STREAMING)) {
		fprintf(stderr, "%s does not support streaming i/o\n", dev_name);
		return 0;
	}

	/* Video standard and tune */
	memset(&cropcap, 0, sizeof(cropcap));
	cropcap.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;

	if(xioctl(fd, VIDIOC_CROPCAP, &cropcap) == 0)
	{
		crop.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
		crop.c = cropcap.defrect; /* reset to default */
		if(xioctl(fd, VIDIOC_S_CROP, &crop) == -1)
		{
			/* Errors ignored. */
		}
	} 
	else 
	{        
		/* Errors ignored. */
	}

	/* Create and submit video format (fmt) */
	memset(&fmt, 0, sizeof(fmt));
	fmt.type                = V4L2_BUF_TYPE_VIDEO_CAPTURE;
	fmt.fmt.pix.width       = width; 
	fmt.fmt.pix.height      = height;
	fmt.fmt.pix.pixelformat = V4L2_PIX_FMT_YUYV; /* V4L2_PIX_FMT_RGB24 or V4L2_PIX_FMT_BGR24 or V4L2_PIX_FMT_YUYV */
	fmt.fmt.pix.field       = V4L2_FIELD_INTERLACED;

	if(xioctl (fd, VIDIOC_S_FMT, &fmt) == -1)
	{
		fprintf(stderr, "ERROR Video format (fmt)");
		return 0;
	}

	/* Buggy driver paranoia. */
	min = fmt.fmt.pix.width * 2;
	if(fmt.fmt.pix.bytesperline < min)
		fmt.fmt.pix.bytesperline = min;
	min = fmt.fmt.pix.bytesperline * fmt.fmt.pix.height;
	if(fmt.fmt.pix.sizeimage < min)
		fmt.fmt.pix.sizeimage = min;

	/* Initiate nmap (memory mapping) */
	memset(&req, 0, sizeof(req));
	req.count    = 3; /* Number of buffers */
	req.type     = V4L2_BUF_TYPE_VIDEO_CAPTURE;
	req.memory   = V4L2_MEMORY_MMAP;

	if(xioctl (fd, VIDIOC_REQBUFS, &req) == -1) 
	{
		fprintf(stderr, "%s does not support memory mapping\n", dev_name);
		return 0;
	}

	if(req.count < 2) 
	{
		fprintf(stderr, "Insufficient buffer memory on %s\n", dev_name);
		return 0;
	}

	buffers = calloc(req.count, sizeof(*buffers));

	if(!buffers) 
	{
		fprintf(stderr, "Out of memory\n");
		return 0;
	}

	/* For each created buffer ... */
	for(n_buffers = 0; n_buffers < req.count; ++n_buffers) 
	{
		struct v4l2_buffer buf;

		memset(&buf, 0, sizeof(buf));
		buf.type    = V4L2_BUF_TYPE_VIDEO_CAPTURE;
		buf.memory  = V4L2_MEMORY_MMAP;
		buf.index   = n_buffers;

		/* Prepare to query buffer into the driver's queue */
		if(xioctl(fd, VIDIOC_QUERYBUF, &buf) == -1)
		{
			fprintf(stderr, "ERROR QUERYBUF\n");
			return 0;
		}

		buffers[n_buffers].length = buf.length;
		buffers[n_buffers].start =
		    mmap(NULL,
		         buf.length,
		         PROT_READ | PROT_WRITE,
		         MAP_SHARED,
		         fd, buf.m.offset);

		if(buffers[n_buffers].start == MAP_FAILED)
		{
			fprintf(stderr, "ERROR NMAP\n");
			return 0;
		}
	}



	/* Start Capturing */
	unsigned int j;
	for (j = 0; j < n_buffers; ++j) {
		struct v4l2_buffer buf;
		memset(&buf, 0, sizeof(buf));
		buf.type        = V4L2_BUF_TYPE_VIDEO_CAPTURE;
		buf.memory      = V4L2_MEMORY_MMAP;
		buf.index       = j;

		/* Add buffer into the drivers buffer queue */
		if (xioctl(fd, VIDIOC_QBUF, &buf) == -1)
		{
			fprintf(stderr, "ERROR QBUF\n");
			return 0;
		}
	}

	/* submit capturing type */
	type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
	if(xioctl(fd, VIDIOC_STREAMON, &type) == -1)
	{
		fprintf(stderr, "ERROR STREAMON\n");
		return 0;
	}

	return 1;
}



/**
 * Reads a frame
 * @return success -> 1  /  failed -> 0
**/
void camcapture_get_frame(image_data_t *frameinfo)
{
	frameinfo->width  = init_width;
	frameinfo->height = init_height;
	frameinfo->img    = (void *) buffers[v4lBuf.index].start;
}



/**
 * Waits for the device to get ready (select) und locks the driver 
 * so he won't modify the buffer
 * If no error occour, the processing function is called
 * @return Index of the buffer that has to be read
**/
int camcapture_lock()
{
	fd_set fds;
	struct timeval tv;
	int r;

	/* loop until select is successfull */
	while(1)
	{
		FD_ZERO (&fds);
		FD_SET (fd, &fds);

		/* Timeout: */
		tv.tv_sec = 2;
		tv.tv_usec = 0;

		r = select(fd+1, &fds, NULL, NULL, &tv);

		if (-1 == r) {
			if (EINTR == errno)
				continue;
			fprintf(stderr, "ERROR SELECT\n");
			exit(EXIT_FAILURE);
		}

		if (0 == r) 
		{
		    fprintf (stderr, "ERROR SELECT TIMEOUT\n");
		    exit(EXIT_FAILURE);
		    return 0;
		}

		/* Set buf */
		memset(&v4lBuf, 0, sizeof(v4lBuf));
		v4lBuf.type   = V4L2_BUF_TYPE_VIDEO_CAPTURE;
		v4lBuf.memory = V4L2_MEMORY_MMAP;

		/* dequeue buffer in the drivers queue (lock) */
		if(xioctl(fd, VIDIOC_DQBUF, &v4lBuf) == -1) 
		{
			if(errno == EAGAIN) 
			{
				continue;
			}
			else
			{
				fprintf(stderr, "ERROR DQBUF");
				exit(EXIT_FAILURE);
			}
		}
		assert(v4lBuf.index < n_buffers);
		break;
	}

	return 1;
}



/**
 * This function unlocks the buffer
 * @return success -> 1  /  failed -> 0
**/
int camcapture_unlock()
{
	/* enqueue buffer in the drivers incoming queue */
	if(xioctl(fd, VIDIOC_QBUF, &v4lBuf) == -1)
	{
		fprintf(stderr, "ERROR QBUF");
		exit(EXIT_FAILURE);
	}
	return 1;
}



/**
 * Destroy, cleanup and free everything
**/
void camcapture_destroy()
{
	/* stop capturing */
	type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
	if(xioctl(fd, VIDIOC_STREAMOFF, &type) == -1)
	{
		fprintf(stderr, "ERROR STREAMOFF\n");
		exit(EXIT_FAILURE);
	}

	/* cleanup nmap buffers */
	unsigned int i;
	for (i = 0; i < n_buffers; ++i)
	{
		if (munmap (buffers[i].start, buffers[i].length) == -1)
		{
			fprintf(stderr, "ERROR NMAP CLEANUP (UNMAP)\n");
			exit(EXIT_FAILURE);
		}
	}
	free(buffers);

	/* close the video device */
	close(fd);
}
